package Orangutan::Contexts;

use strict;

use Data::Dumper;
use Orangutan::Context;

sub new {
    my $class = shift;
    my $self = { };
    $self->{'contexts'} = { };
    $self->{'contexts'}->{'in'} = { };
    $self->{'contexts'}->{'out'} = { };
    $self->{'contexts'}->{'handlers'} = { };
    $self->{'undo'} = undef;
    $self->{'stack'} = [ ];
    bless $self, $class;
    return $self;
}

sub AddRequestResponse($$) {
    my ($self, $context, $outweight, $inweight) = @_;
    my $id = $context->GetID;

    if ($context->ShouldCopy) {
        # Try to reuse already added context
        if (defined($self->{'contexts'}->{'out'}->{$id})) {
            $context = $self->{'contexts'}->{'out'}->{$id}[1];
        } elsif (defined($self->{'contexts'}->{'in'}->{$id})) {
            $context = $self->{'contexts'}->{'in'}->{$id}[1];
        } else {
            $context = new Orangutan::Context($context);
        }
        if ($context->HasDelay) { # ShouldCopy calls HasDelay
            $context->InitDelay;
        }
    }

    if (!defined($outweight)) {
        $outweight = 0;
    }
    if (!defined($inweight)) {
        $inweight = 0;
    }

    if (defined($self->{'contexts'}->{'out'}->{$id}) &&
        defined($self->{'contexts'}->{'in'}->{$id})) {
        $main::logger->Log("context %s has been already added", $id);
    } else {
        if (!defined($self->{'contexts'}->{'out'}->{$id})) {
            $self->{'contexts'}->{'out'}->{$id} = [ $outweight, $context ];
        } else {
            $main::logger->Log("output context %s has been already added", $id);
        }
        if (!defined($self->{'contexts'}->{'in'}->{$id})) {
            $self->{'contexts'}->{'in'}->{$id} = [ $inweight, $context ];
        } else {
            $main::logger->Log("input context %s has been already added", $id);
        }
    }

    return $context;
}

sub AddRequest($$) {
    my ($self, $context, $weight) = @_;

    if ($context->ShouldCopy) {
        $context = new Orangutan::Context($context);
        if ($context->HasDelay) {
            $context->InitDelay;
        }
    }

    if (!defined($weight)) {
        $weight = 0;
    }

    my $id = $context->GetID;
    if (!defined($self->{'contexts'}->{'out'}->{$id})) {
        $self->{'contexts'}->{'out'}->{$id} = [ $weight, $context ];
    } else {
        $main::logger->Log("output context %s has been already added", $id);
    }

    return $context;
}

sub AddResponse($$) {
    my ($self, $context, $weight) = @_;

    if ($context->ShouldCopy) {
        $context = new Orangutan::Context($context);
    }

    if (!defined($weight)) {
        $weight = 0;
    }

    if ($context->ShouldTell) {
        $context->Activate;
    }

    my $id = $context->GetID;
    if (!defined($self->{'contexts'}->{'in'}->{$id})) {
        $self->{'contexts'}->{'in'}->{$id} = [ $weight, $context ];
    } else {
        $main::logger->Log("input context %s has been already added", $id);
    }

    return $context;
}

sub AddHandler($$) {
    my ($self, $context) = @_;

    my $id = $context->GetID;
    foreach my $event (keys %{$context->{'event'}}) {
        if (!defined($self->{'contexts'}->{'handlers'}->{$event})) {
            $self->{'contexts'}->{'handlers'}->{$event} = { };
        }
        if (!defined($self->{'contexts'}->{'handlers'}->{$event}->{$id})) {
            $self->{'contexts'}->{'handlers'}->{$event}->{$id} = $context;
        } else {
            $main::logger->Log("context %s has been already added", $id);
        }
    }
}

sub SetUndo($;$) {
    my ($self, $context) = @_;
    $self->{'undo'} = $context;
}

sub RemoveRequestResponse($$) {
    my ($self, $context) = @_;
    $self->Remove('out', $context);
    $self->Remove('in', $context);
}

sub RemoveRequest($$) {
    my ($self, $context) = @_;
    $self->Remove('out', $context);
}

sub RemoveResponse($$) {
    my ($self, $context) = @_;
    $self->Remove('in', $context);
}

sub Remove($$$) {
    my ($self, $type, $context) = @_;
    delete($self->{'contexts'}->{$type}->{$context->GetID});
}

sub GetRequest($$) {
    my ($self, $id) = @_;
    if (defined($self->{'contexts'}->{'out'}->{$id})) {
        return $self->{'contexts'}->{'out'}->{$id}[1];
    } else {
        return undef;
    }
}

sub GetResponse($$) {
    my ($self, $id) = @_;
    if (defined($self->{'contexts'}->{'in'}->{$id})) {
        return $self->{'contexts'}->{'in'}->{$id}[1];
    } else {
        return undef;
    }
}

sub GetTimeout($) {
    my $self = shift;

    foreach my $id (keys %{$self->{'contexts'}->{'in'}}) {
        if (defined($self->{'contexts'}->{'in'}->{$id})) {
            my $context = $self->{'contexts'}->{'in'}->{$id}[1];
            my $timeout = $context->GetTimeout;
            if (defined($timeout) && ($timeout > 0) && (($timeout - time) <= 0)) {
                my $keep = undef;
                if ($context->HasTimeoutHandler) {
                    if (!($keep = $context->CallTimeoutHandler($self))) {
                        $timeout = $context->GetTimeout;
                    } else {
                        $timeout = undef;
                    }
                    $self->SetUndo($context);
                }
                if (!$keep) {
                    $self->RemoveResponse($context);
                    $timeout = undef;
                }
            }
            if (defined($timeout) && ($timeout > 0)) {
                return $timeout;
            }
        }
    }

    return 0;
}

sub GetDelay($) {
    my $self = shift;

    my $delay = 0;
    foreach my $id (keys %{$self->{'contexts'}->{'out'}}) {
        my $context = $self->{'contexts'}->{'out'}->{$id}[1];
        my $time = $context->GetDelay;
        if (defined($time) && (($delay == 0) || ($time < $delay))) {
            $delay = $time;
        }
    }

    return $delay;
}

sub GetStack($) {
    my $self = shift;
    return @{$self->{'stack'}};
}

sub UpdateStack($$) {
    my ($self, $context) = @_;
    unshift(@{$self->{'stack'}}, $context->GetID);
    if ($#{$self->{'stack'}} == 10) {
        pop(@{$self->{'stack'}});
    }
}

sub Undo($) {
    my $self = shift;
    if (defined($self->{'undo'})) {
        if ($self->{'undo'}->HasUndo) {
            $self->{'undo'}->CallUndo($self);
            $self->SetUndo;
            return 1;
        }
    }
    return 0;
}

sub Suspend($) {
    my $self = shift;

    my $time = time;

    foreach my $id (keys %{$self->{'contexts'}->{'in'}}) {
        my $context = $self->{'contexts'}->{'in'}->{$id}[1];
        my $timeout = $context->GetTimeout;
        if (defined($timeout) && ($timeout > $time)) {
            $context->SetTimeout($time - $timeout);
        }
        if ($context->HasForeignHandlers) {
            $context->SuspendForeignHandlers;
        }
    }

    foreach my $id (keys %{$self->{'contexts'}->{'out'}}) {
        my $context = $self->{'contexts'}->{'out'}->{$id}[1];
        my $delay = $context->GetDelay;
        if (defined($delay) && ($delay > 0) && ($delay > $time)) {
            $context->SetDelay($time - $delay);
        }
    }
}

sub Resume($) {
    my $self = shift;

    my $time = time;

    foreach my $id (keys %{$self->{'contexts'}->{'in'}}) {
        my $context = $self->{'contexts'}->{'in'}->{$id}[1];
        my $timeout = $context->GetTimeout;
        if (defined($timeout) && ($timeout < 0)) {
            $context->SetTimeout($time - $timeout);
        }
        if ($context->HasForeignHandlers) {
            $context->ResumeForeignHandlers;
        }
    }

    foreach my $id (keys %{$self->{'contexts'}->{'out'}}) {
        my $context = $self->{'contexts'}->{'out'}->{$id}[1];
        my $delay = $context->GetDelay;
        if (defined($delay) && ($delay < 0)) {
            $context->SetDelay($time - $delay);
        }
    }
}

sub Run($) {
    my $self = shift;

    foreach my $id (sort {
            $self->{'contexts'}->{'out'}->{$a}[0] <=> $self->{'contexts'}->{'out'}->{$b}[0]
        } keys %{$self->{'contexts'}->{'out'}}) {
        my $context = $self->{'contexts'}->{'out'}->{$id}[1];
        my $delay = $context->GetDelay;
        if (!defined($delay) || (($delay - time) <= 0)) {
            $context->Tell($self);
            $self->RemoveRequest($context);
            if ($context->ShouldWait) {
                $context->InitTimeout;
                return $context->GetTimeout;
            }
        }
    }

    return 0;
}

sub Match($$) {
    my ($self, $message) = @_;

    foreach my $id (sort {
            $self->{'contexts'}->{'in'}->{$a}[0] <=> $self->{'contexts'}->{'in'}->{$b}[0]
        } keys %{$self->{'contexts'}->{'in'}}) {
        my $context = $self->{'contexts'}->{'in'}->{$id}[1];
        if ($context->IsActive) {
            my @matches = $context->Hear($message);
            if (scalar @matches > 0) {
                my $keep = undef;
                if ($context->HasHandler) {
                    $keep = $context->CallHandler($self, @matches);
                }
                if ($context->ShouldReply) {
                    $context->Reply($self);
                }
                if (!$context->IsReusable && !$keep) {
                    $self->RemoveResponse($context);
                }
                $self->SetUndo($context);
                return;
            }
        }
    }

    foreach my $id (keys %{$self->{'contexts'}->{'in'}}) {
        my $context = $self->{'contexts'}->{'in'}->{$id}[1];
        if ($context->IsActive && $context->HasNoMatchHandler) {
            $context->CallNoMatchHandler($self);
        }
    }
}

1;
