package Orangutan::Query;

use strict;

use DBI;
use Orangutan::Date;

sub new {
    my $class = shift;

    my $self = { };

    my %args = ( );
    while ($#_ >= 0) {
        $args{lc(pop)} = pop;
    }

    bless $self, $class;

    my ($hostname, $database, $timeout);
    foreach my $arg (keys %args) {
        if (($arg eq 'host') ||
            ($arg eq 'hostname')) {
            $hostname = $args{$arg};
        } elsif
           (($arg eq 'db') ||
            ($arg eq 'database')) {
            $database = $args{$arg};
        } elsif
           (($arg eq 'user') ||
            ($arg eq 'username')) {
            $self->{'username'} = $args{$arg};
        } elsif
           (($arg eq 'pass') ||
            ($arg eq 'password')) {
            $self->{'password'} = $args{$arg};
        } elsif ($arg eq 'timeout') {
            $timeout = $args{$arg};
        } else {
            $main::logger->Log("unsupported query property: %s", $arg);
        }
    }

    if (defined($database)) {
        $self->{'dsn'} = "DBI:mysql:database=$database";
        if (defined($hostname)) {
            $self->{'dsn'} .= ";host=$hostname";
        }
        if (defined($timeout)) {
            $self->{'dsn'} .= ";mysql_connect_timeout=$timeout";
        }
        $self->{'dsn'} .= ";mysql_enable_utf8=1";
        return $self;
    } else {
        return undef;
    }
}

sub Connect($) {
    my $self = shift;
    if ($self->{'dbh'}->{'Active'}) {
        $self->{'keep'}++;
        return $self->{'dbh'};
    }
    if (($self->{'dbh'} = DBI->connect($self->{'dsn'}, $self->{'username'}, $self->{'password'}))) {
        return $self->{'dbh'};
    } else {
        $main::logger->Log("could not connect to database");
        return undef;
    }
}

sub Escape($$) {
    my ($self, $string) = @_;
    if ($self->{'dbh'}) {
        return $self->{'dbh'}->quote($string);
    }
    return 'NULL';
}

sub Query($$) {
    my ($self, $sql) = @_;
    if ($self->{'dbh'}) {
        $self->{'sth'} = $self->{'dbh'}->prepare($sql);
        if ($self->{'sth'}) {
            my $result = $self->{'sth'}->execute;
            if ($result) {
                if (defined($self->{'sth'}->{'NUM_OF_FIELDS'}) &&
                    ($self->{'sth'}->{'NUM_OF_FIELDS'} > 0)) {
                    $self->{'results'} = [ ];
                    while (defined(my $hashref = $self->{'sth'}->fetchrow_hashref)) {
                        push(@{$self->{'results'}}, $hashref);
                    }
                    if (scalar @{$self->{'results'}} > 0) {
                        $self->{'index'} = 0;
                    }
                } else {
                    $self->{'results'} = $result;
                }
                return $self->{'results'};
            } else {
                $main::logger->Log("failed to execute sql statement: %s", $self->{'sth'}->errstr);
            }
        } else {
            $main::logger->Log("could not prepare sql statement: %s", $self->{'dbh'}->errstr);
        }
    }
}

sub Rows($) {
    my $self = shift;
    return scalar @{$self->{'results'}};
}

sub First($) {
    my $self = shift;
    $self->{'index'} = 0;
    return $self->Next;
}

sub Next($) {
    my $self = shift;
    return $self->{'results'}[$self->{'index'}++];
}

sub GetRow($$) {
    my ($self, $index) = @_;
    return $self->{'results'}[$index];
}

sub Last($) {
    my $self = shift;
    $self->{'index'} = $#{$self->{'results'}};
    return $self->Next;
}

sub Disconnect($) {
    my $self = shift;
    if ($self->{'dbh'}) {
        if ($self->{'keep'}) {
            $self->{'keep'}--;
        } else {
            $self->{'dbh'}->disconnect;
        }
    }
}

sub UserExists($$) {
    my ($self, $username) = @_;

    my $result = undef;
    if ($self->Connect) {
        $result = '';
        $username = $self->Escape($username);
        if ($self->Query(qq{
            SELECT id
            FROM users
            WHERE login = $username AND status = 1 AND type = 'User'
            })) {
            if ($self->Rows > 0) {
                $result = $self->Next->{'id'};
            }
        }
        $self->Disconnect;
    }
    return $result;
}

sub ProjectExists($$) {
    my ($self, $name) = @_;

    my $projects = undef;
    if ($self->Connect) {
        $projects = '';
        my $project = $self->Escape($name);
        if ($self->Query(qq{
            SELECT id, name
            FROM projects
            WHERE name = $project AND status = 1
            })) {
            if ($self->Rows == 1) {
                my $result = $self->Next;
                $projects = {
                    id   => $result->{'id'},
                    name => $result->{'name'}
                };
            } elsif ($self->Rows > 1) {
                $projects = [ ];
                while (defined(my $result = $self->Next)) {
                    push(@{$projects}, {
                        id   => $result->{'id'},
                        name => $result->{'name'}
                    });
                }
            }
        }
        $self->Disconnect;
    }
    return $projects;
}

sub ActivityExists($$;$) {
    my ($self, $name, $projectid) = @_;

    my $activity = undef;
    if ($self->Connect) {
        my $sql = '';
        $activity = '';
        $name = $self->Escape($name);
        if (defined($projectid)) {
            $sql = qq{
                SELECT IFNULL(child.id, parent.id) AS id, parent.name
                FROM enumerations parent
                LEFT JOIN enumerations child ON child.parent_id = parent.id AND child.project_id = $projectid
                WHERE parent.name = $name AND parent.type = 'TimeEntryActivity' AND parent.parent_id IS NULL AND
                ((child.active IS NULL AND parent.active = 1) OR (child.active IS NOT NULL AND child.active = 1))
            };
        } else {
            $sql = qq{
                SELECT MIN(id) AS id, name
                FROM enumerations
                WHERE name = $name AND type = 'TimeEntryActivity' AND active = 1
                GROUP BY name
            };
        }
        if ($self->Query($sql)) {
            if ($self->Rows > 0) {
                $activity = $self->Next->{'id'};
            }
        }
        $self->Disconnect;
    }
    return $activity;
}

sub CanView($$$) {
    my ($self, $user, $projectid) = @_;
    my $result = 0;
    if ($self->Connect) {
        my $employeeid = $user->GetUserID;
        if ($self->Query(qq{
            SELECT role_id
            FROM members
            JOIN member_roles ON member_id = members.id
            JOIN roles ON roles.id = role_id
            WHERE user_id = $employeeid AND project_id = $projectid AND permissions LIKE '%- :view_issues\n%';
            })) {
            $result = $self->Rows;
        }
        $self->Disconnect;
    }
    return $result;
}

sub CanLog($$$) {
    my ($self, $user, $projectid) = @_;
    my $result = 0;
    if ($self->Connect) {
        my $employeeid = $user->GetUserID;
        if ($self->Query(qq{
            SELECT role_id
            FROM members
            JOIN member_roles ON member_id = members.id
            JOIN roles ON roles.id = role_id
            WHERE user_id = $employeeid AND project_id = $projectid AND permissions LIKE '%- :edit_time_entries\n%';
            })) {
            $result = $self->Rows;
        }
        $self->Disconnect;
    }
    return $result;
}

sub GetProjects($;$) {
    my ($self, $projectid) = @_;
    my $rows = undef;
    if ($self->Connect) {
        my $and = '';
        if (defined($projectid)) {
            $and = qq{ AND project.parent_id = $projectid };
        } else {
            $and = qq{ AND project.parent_id IS NULL };
        }
        if ($self->Query(qq{
            SELECT project.name AS project, parent.name AS parent,
            (SELECT COUNT(*) FROM projects WHERE parent_id = project.id) AS subprojects,
            UNIX_TIMESTAMP(MAX(time_entries.spent_on)) AS end_time
            FROM projects project
            LEFT JOIN projects parent ON parent.id = project.parent_id
            LEFT JOIN time_entries ON time_entries.project_id = project.id
            WHERE project.status = 1
            $and
            GROUP BY project
            ORDER BY project
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetTrackers($) {
    my $self = shift;
    my $rows = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT id, name
            FROM trackers
            ORDER BY name
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetStatuses($) {
    my $self = shift;
    my $rows = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT id, name
            FROM issue_statuses
            ORDER BY name
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetPriorities($) {
    my $self = shift;
    my $rows = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT id, name
            FROM enumerations
            WHERE type = 'IssuePriority' AND parent_id IS NULL AND active = 1
            ORDER BY name
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetIssue($$) {
    my ($self, $issueid) = @_;
    my $issue = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT issues.id, tracker_id, issues.project_id, subject, issues.description,
            due_date, category_id, status_id, issues.assigned_to_id, priority_id, author_id,
            issues.created_on, issues.updated_on, start_date, done_ratio, estimated_hours,
            trackers.name AS tracker_name, projects.name AS project_name,
            issue_categories.name AS category_name, issue_statuses.name AS status_name, is_closed,
            users.login AS assigned_to_name, enumerations.name AS priority_name,
            authors.login AS author_name
            FROM issues
            JOIN trackers ON trackers.id = tracker_id
            JOIN projects ON projects.id = project_id
            LEFT JOIN issue_categories ON issue_categories.id = category_id
            JOIN issue_statuses ON issue_statuses.id = status_id
            LEFT JOIN users users ON users.id = issues.assigned_to_id
            JOIN enumerations ON enumerations.id = priority_id
            JOIN users authors ON authors.id = author_id
            WHERE issues.id = $issueid
            })) {
            if ($self->Rows > 0) {
                $issue = $self->Next;
            }
        }
        $self->Disconnect;
    }
    return $issue;
}

sub GetIssueEx($$) {
    my ($self, $issueid) = @_;
    my $issue = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT issues.id, issues.parent_id,
            (SELECT COUNT(*) FROM issues childs WHERE childs.parent_id = issues.id) AS childs,
            trackers.name AS tracker_name, issues.project_id, projects.name AS project_name,
            subject, issues.description, UNIX_TIMESTAMP(due_date) AS due_date, issue_categories.name AS category_name,
            issue_statuses.name AS status_name, is_closed, users.login AS assigned_to_name,
            enumerations.name AS priority_name, authors.login AS author_name,
            UNIX_TIMESTAMP(issues.created_on) AS created_on, UNIX_TIMESTAMP(issues.updated_on) AS updated_on,
            UNIX_TIMESTAMP(start_date) AS start_date, done_ratio, (estimated_hours * 3600) AS estimated_hours,
            (SELECT SUM(hours * 3600) FROM time_entries WHERE issue_id = issues.id) AS spent_time,
            (SELECT COUNT(*) FROM journals WHERE journalized_type = 'Issue' AND journalized_id = issues.id
            AND notes IS NOT NULL AND notes != '') AS notes,
            (SELECT COUNT(*) FROM votes WHERE voteable_type = 'Issue' AND voteable_id = issues.id) AS votes
            FROM issues
            JOIN trackers ON trackers.id = tracker_id
            JOIN projects ON projects.id = issues.project_id
            LEFT JOIN issue_categories ON issue_categories.id = category_id
            JOIN issue_statuses ON issue_statuses.id = status_id
            LEFT JOIN users users ON users.id = issues.assigned_to_id
            JOIN enumerations ON enumerations.id = priority_id
            LEFT JOIN versions ON versions.id = fixed_version_id
            JOIN users authors ON authors.id = author_id
            WHERE issues.id = $issueid
            })) {
            if ($self->Rows > 0) {
                $issue = $self->Next;
            }
        }
        $self->Disconnect;
    }
    return $issue;
}

sub GetBestIssues($$;$$) {
    my ($self, $user, $projectid, $date) = @_;
    my $rows = undef;
    if ($self->Connect) {
        my $and = '';
        my $employeeid = $user->GetUserID;
        if (defined($projectid)) {
            $and = qq{ AND issues.project_id = $projectid };
        }
        if (!defined($date)) {
            $date = new Orangutan::Date;
        }
        my $currentdate = $self->Escape($date->DateTime('date'));
        if ($self->Query(qq{
            SELECT issues.id, trackers.name AS tracker_name, projects.name AS project_name,
            subject, issue_statuses.name AS status_name, enumerations.name AS priority_name
            FROM issues
            JOIN trackers ON trackers.id = tracker_id
            JOIN projects ON projects.id = project_id
            JOIN issue_statuses ON issue_statuses.id = status_id
            JOIN enumerations ON enumerations.id = priority_id
            JOIN members ON members.project_id = issues.project_id
            JOIN member_roles ON member_id = members.id
            JOIN roles ON roles.id = role_id
            WHERE user_id = $employeeid AND permissions LIKE '%- :edit_time_entries\n%'
            AND (assigned_to_id IS NULL OR assigned_to_id = $employeeid)
            AND done_ratio < 100 AND is_closed = 0
            AND (start_date IS NULL OR start_date <= $currentdate)
            $and
            GROUP BY issues.id
            ORDER BY assigned_to_id DESC,
            DATEDIFF($currentdate, IFNULL(due_date, $currentdate)) DESC,
            enumerations.position DESC, done_ratio, issue_statuses.position, issues.id
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetIssuesEx($) {
    my ($self, $user, $userid, $projectid, $trackerid, $statusid, $priorityid, %options) = @_;
    my $rows = undef;
    if ($self->Connect) {
        my @clause = ( );
        if (!defined($userid)) {
            if (!$user->IsCreator && !$user->IsManager) {
                $userid = $user->GetUserID;
                push(@clause, qq{ user_id = $userid });
                push(@clause, qq{ permissions LIKE '%- :view_issues\n%' });
            }
            push(@clause, qq{ assigned_to_id IS NULL });
        } elsif ($userid eq '') {
            $userid = $user->GetUserID;
            if (!$user->IsCreator && !$user->IsManager) {
                push(@clause, qq{ user_id = $userid });
                push(@clause, qq{ permissions LIKE '%- :view_issues\n%' });
            }
            if (!defined($options{'all'})) {
                push(@clause, qq{ (assigned_to_id = $userid OR assigned_to_id IS NULL) });
            }
        } else {
            push(@clause, qq{ user_id = $userid });
            push(@clause, qq{ permissions LIKE '%- :view_issues\n%' });
            push(@clause, qq{ assigned_to_id = $userid });
        }
        if (defined($projectid)) {
            push(@clause, qq{ issues.project_id = $projectid });
        }
        if (defined($trackerid)) {
            push(@clause, qq{ tracker_id = $trackerid });
        }
        if (!defined($statusid)) {
            push(@clause, qq{ is_closed = 0 });
        } elsif ($statusid eq '') {
            push(@clause, qq{ is_closed = 1 });
        } else {
            push(@clause, qq{ status_id = $statusid });
        }
        if (defined($priorityid)) {
            push(@clause, qq{ priority_id = $priorityid });
        }
        if (defined($options{'actual'}) || defined($options{'overdue'})) {
            my $date = new Orangutan::Date;
            my $currentdate = $self->Escape($date->DateTime('date'));
            if (defined($options{'actual'})) {
                push(@clause, qq{ (start_date IS NULL OR start_date <= $currentdate) });
                push(@clause, qq{ (due_date IS NULL OR due_date >= $currentdate) });
            } else {
                push(@clause, qq{ due_date IS NOT NULL });
                push(@clause, qq{ due_date < $currentdate });
            }
        }
        if (defined($options{'overtimed'})) {
            push(@clause, qq{ estimated_hours IS NOT NULL });
            push(@clause, qq{ (SELECT SUM(hours) FROM time_entries WHERE issue_id = issues.id) > estimated_hours });
        }
        my $join = '';
        if (defined($userid)) {
            $join = qq{
                JOIN members ON members.project_id = issues.project_id
                JOIN member_roles ON member_id = members.id
                JOIN roles ON roles.id = role_id
            };
        }
        my $where = '';
        if (scalar @clause > 0) {
            $where = 'WHERE ';
            $where .= join(' AND ', @clause);
        }
        if ($self->Query(qq{
            SELECT issues.id, trackers.name AS tracker_name, projects.name AS project_name,
            subject, issue_statuses.name AS status_name, enumerations.name AS priority_name
            FROM issues
            JOIN trackers ON trackers.id = tracker_id
            JOIN projects ON projects.id = project_id
            JOIN issue_statuses ON issue_statuses.id = status_id
            JOIN enumerations ON enumerations.id = priority_id
            $join
            $where
            GROUP BY issues.id
            ORDER BY enumerations.position DESC, issues.id
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetRelatedIssues($$) {
    my ($self, $issueid) = @_;
    my $rows = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT relation_type, issue_from_id, issue_to_id, name AS tracker_name, subject, delay
            FROM issue_relations
            JOIN issues ON issues.id = issue_to_id
            JOIN trackers ON trackers.id = tracker_id
            WHERE issue_from_id = $issueid OR issue_to_id = $issueid
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetSubissues($$) {
    my ($self, $issueid) = @_;
    my $rows = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT issues.id, trackers.name AS tracker_name, projects.name AS project_name,
            subject, issue_statuses.name AS status_name, enumerations.name AS priority_name
            FROM issues
            JOIN trackers ON trackers.id = tracker_id
            JOIN projects ON projects.id = project_id
            JOIN issue_statuses ON issue_statuses.id = status_id
            JOIN enumerations ON enumerations.id = priority_id
            WHERE issues.parent_id = $issueid
            ORDER BY issues.id
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetVote($$) {
    my ($self, $issueid) = @_;
    my $vote = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT SUM(IF(vote = 0, -1, 1)) AS vote, SUM(vote) AS yes, SUM(IF(vote = 0, 1, 0)) AS no
            FROM votes
            WHERE voteable_type = 'Issue' AND voteable_id = $issueid
            })) {
            if ($self->Rows > 0) {
                $vote = $self->Next;
            }
        }
        $self->Disconnect;
    }
    return $vote;
}

sub GetAttachments($$) {
    my ($self, $issueid) = @_;
    my $rows = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT id, filename, filesize, description
            FROM attachments
            WHERE container_type = 'Issue' AND container_id = $issueid
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetWatchers($$) {
    my ($self, $issueid) = @_;
    my $rows = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT login
            FROM watchers
            JOIN users ON users.id = user_id
            WHERE watchable_type = 'Issue' AND watchable_id = $issueid
            ORDER BY login
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetDefaultActivity($$) {
    my ($self, $projectid) = @_;
    my $activity = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT IFNULL(child.id, parent.id) AS id, parent.name
            FROM enumerations parent
            LEFT JOIN enumerations child ON child.parent_id = parent.id AND child.project_id = $projectid
            WHERE parent.type = 'TimeEntryActivity' AND parent.parent_id IS NULL AND
            ((child.active IS NULL AND parent.active = 1) OR (child.active IS NOT NULL AND child.active = 1))
            ORDER BY parent.is_default DESC, parent.position
            LIMIT 1
            })) {
            if ($self->Rows > 0) {
                $activity = $self->Next;
            }
        }
        $self->Disconnect;
    }
    return $activity;
}

sub GetActivities($$) {
    my ($self, $projectid) = @_;
    my $rows = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT IFNULL(child.id, parent.id) AS id, parent.name
            FROM enumerations parent
            LEFT JOIN enumerations child ON child.parent_id = parent.id AND child.project_id = $projectid
            WHERE parent.type = 'TimeEntryActivity' AND parent.parent_id IS NULL AND
            ((child.active IS NULL AND parent.active = 1) OR (child.active IS NOT NULL AND child.active = 1))
            ORDER BY parent.position
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetLastEntry($$;$) {
    my ($self, $user, $date) = @_;

    my $event = undef;
    if ($self->Connect) {
        $event = '';
        my $and = '';
        my $employeeid = $user->GetUserID;
        if (defined($date)) {
            my $start_time = $self->Escape($date->DateTime);
            $and = qq{ AND start_time < $start_time };
        }
        if ($self->Query(qq{
            SELECT time_entries.id, time_entries.project_id, user_id, issue_id, comments, activity_id,
            UNIX_TIMESTAMP(start_time) AS start_time, UNIX_TIMESTAMP(end_time) AS end_time,
            projects.name AS project, enumerations.name AS activity
            FROM time_entries
            LEFT JOIN projects ON projects.id = time_entries.project_id
            LEFT JOIN enumerations ON enumerations.id = activity_id
            WHERE user_id = $employeeid AND start_time IS NOT NULL AND end_time IS NOT NULL $and
            ORDER BY end_time DESC
            LIMIT 1
            })) {
            if ($self->Rows > 0) {
                $event = $self->Next;
            }
        }
        $self->Disconnect;
    }
    return $event;
}

sub GetEntries($$$$) {
    my ($self, $user, $start, $end) = @_;
    my $rows = undef;
    if ($self->Connect) {
        my $employeeid = $user->GetUserID;
        my $startdate = $self->Escape($start->DateTime);
        my $enddate = $self->Escape($end->DateTime);
        if ($self->Query(qq{
            SELECT time_entries.id, time_entries.project_id, user_id, issue_id, comments, activity_id,
            UNIX_TIMESTAMP(start_time) AS start_time, UNIX_TIMESTAMP(end_time) AS end_time,
            projects.name AS project, enumerations.name AS activity
            FROM time_entries
            LEFT JOIN projects ON projects.id = time_entries.project_id
            LEFT JOIN enumerations ON enumerations.id = activity_id
            WHERE user_id = $employeeid AND start_time IS NOT NULL AND end_time IS NOT NULL
            AND start_time < $enddate AND end_time > $startdate
            ORDER BY start_time
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetEntriesEx($$$$$$$) {
    my ($self, $userid, $start, $end, $projectid, $issue, $activityid) = @_;
    my $rows = undef;
    if ($self->Connect) {
        my $and = '';
        my $employeeid = $userid;
        my $startdate = $self->Escape($start->DateTime);
        my $enddate = $self->Escape($end->DateTime);
        if (defined($projectid)) {
            $and .= qq{ AND time_entries.project_id = $projectid };
        }
        if (defined($issue)) {
            $and .= qq{ AND issue_id = $issue };
        }
        if (defined($activityid)) {
            if (defined($projectid)) {
                $and .= qq{ AND activity_id = $activityid };
            } else {
                $and .= qq{ AND (activity_id = $activityid OR enumerations.parent_id = $activityid ) };
            }
        }
        if ($self->Query(qq{
            SELECT time_entries.id, time_entries.project_id, user_id, issue_id, comments, activity_id,
            UNIX_TIMESTAMP(start_time) AS start_time, UNIX_TIMESTAMP(end_time) AS end_time,
            CASE WHEN start_time IS NULL AND end_time IS NULL THEN (hours * 3600) ELSE TIMESTAMPDIFF(SECOND,
            CASE WHEN start_time < $startdate THEN $startdate ELSE start_time END,
            CASE WHEN end_time > $enddate THEN $enddate ELSE end_time END) END AS duration,
            projects.name AS project, enumerations.name AS activity,
            CASE WHEN start_time IS NOT NULL THEN UNIX_TIMESTAMP(start_time)
            ELSE UNIX_TIMESTAMP(spent_on) END AS order_time
            FROM time_entries
            LEFT JOIN projects ON projects.id = time_entries.project_id
            LEFT JOIN enumerations ON enumerations.id = activity_id
            WHERE user_id = $employeeid
            AND ((start_time IS NOT NULL AND end_time IS NOT NULL
            AND start_time < $enddate AND end_time > $startdate)
            OR (start_time IS NULL AND end_time IS NULL
            AND DATE_ADD(spent_on, INTERVAL 6 HOUR) >= $startdate
            AND DATE_ADD(spent_on, INTERVAL 6 HOUR) < $enddate))
            $and
            ORDER BY order_time
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetWorkHours($$$$) {
    my ($self, $user, $start, $end) = @_;
    my $hours = undef;
    if ($self->Connect) {
        my $employeeid = $user->GetUserID;
        my $starttime = $self->Escape($start->DateTime);
        my $startdate = $self->Escape($start->DateTime('date'));
        my $endtime = $self->Escape($end->DateTime);
        my $enddate = $self->Escape($end->DateTime('date'));
        if ($self->Query(qq{
            SELECT SUM(
            CASE WHEN start_time IS NULL AND end_time IS NULL THEN (hours * 3600) ELSE TIMESTAMPDIFF(SECOND,
            CASE WHEN start_time < $startdate THEN $startdate ELSE start_time END,
            CASE WHEN end_time > $enddate THEN $enddate ELSE end_time END) END) AS duration
            FROM time_entries
            WHERE user_id = $employeeid
            AND ((start_time IS NOT NULL AND end_time IS NOT NULL
            AND start_time < $enddate AND end_time > $startdate)
            OR (start_time IS NULL AND end_time IS NULL
            AND DATE_ADD(spent_on, INTERVAL 6 HOUR) >= $startdate
            AND DATE_ADD(spent_on, INTERVAL 6 HOUR) < $enddate))
            })) {
            if ($self->Rows > 0) {
                $hours = $self->Next->{'duration'};
            }
        }
        $self->Disconnect;
    }
    return $hours;
}

sub GetWorkHoursEx($$$$$$) {
    my ($self, $userid, $start, $end, $projectid, $activityid) = @_;
    my $rows = undef;
    if ($self->Connect) {
        my $and = '';
        my $employeeid = $userid;
        my $startdate = $self->Escape($start->DateTime);
        my $enddate = $self->Escape($end->DateTime);
        if (defined($projectid)) {
            $and .= qq{ AND time_entries.project_id = $projectid };
        }
        if (defined($activityid)) {
            $and .= qq{ AND activity_id = $activityid };
        }
        if ($self->Query(qq{
            SELECT time_entries.project_id, activity_id,
            projects.name AS project, enumerations.name AS activity,
            SUM(CASE WHEN start_time IS NULL AND end_time IS NULL THEN (hours * 3600) ELSE TIMESTAMPDIFF(SECOND,
            CASE WHEN start_time < $startdate THEN $startdate ELSE start_time END,
            CASE WHEN end_time > $enddate THEN $enddate ELSE end_time END) END) AS duration
            FROM time_entries
            LEFT JOIN projects ON projects.id = time_entries.project_id
            LEFT JOIN enumerations ON enumerations.id = activity_id
            WHERE user_id = $employeeid
            AND ((start_time IS NOT NULL AND end_time IS NOT NULL
            AND start_time < $enddate AND end_time > $startdate)
            OR (start_time IS NULL AND end_time IS NULL
            AND DATE_ADD(spent_on, INTERVAL 6 HOUR) >= $startdate
            AND DATE_ADD(spent_on, INTERVAL 6 HOUR) < $enddate))
            $and
            GROUP BY time_entries.project_id, activity_id
            ORDER BY time_entries.project_id, activity_id
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetLeaveDays($$$$) {
    my ($self, $user, $start, $end) = @_;
    my $rows = undef;
    if ($self->Connect) {
        my $projectid = $main::config->Get('redmine', 'dayoffs_project_id');
        my $employeeid = $user->GetUserID;
        my $startdate = $self->Escape($start->DateTime('date'));
        my $enddate = $self->Escape($end->DateTime('date'));
        if ($self->Query(qq{
            SELECT UNIX_TIMESTAMP(start_date) AS start_date, UNIX_TIMESTAMP(due_date) AS end_date
            FROM issues
            WHERE project_id = $projectid AND assigned_to_id = $employeeid
            AND start_date <= $enddate AND due_date >= $startdate
            ORDER BY start_date
            })) {
            $rows = $self->Rows;
        }
        $self->Disconnect;
    }
    return $rows;
}

sub GetAvgWorkTime {
    my ($self, $user) = @_;
    my @worktime = ( );
    if ($self->Connect) {
        my $employeeid = $user->GetUserID;
        my $weeksago = new Orangutan::Date(time - 1209600);
        my $start_time = $self->Escape($weeksago->WorkDayStart->DateTime);
        if ($self->Query(qq{
            SELECT SEC_TO_TIME(AVG(start_time)) AS start_time, SEC_TO_TIME(AVG(end_time)) AS end_time
            FROM (SELECT MIN(TIME_TO_SEC(TIME(start_time))) AS start_time, MAX(TIME_TO_SEC(TIME(end_time))) AS end_time
            FROM time_entries
            WHERE user_id = $employeeid AND start_time >= $start_time
            AND DATE(start_time) = DATE(end_time) AND TIME(start_time) > '06:00:00'
            GROUP BY DATE(start_time)) AS worktime
            })) {
            if ($self->Rows > 0) {
                my $time = $self->Next;
                if (defined($time->{'start_time'})) {
                    push(@worktime, $time->{'start_time'});
                } else {
                    push(@worktime, undef);
                }
                if (defined($time->{'end_time'})) {
                    push(@worktime, $time->{'end_time'});
                } else {
                    push(@worktime, undef);
                }
            }
        }
        $self->Disconnect;
    }
    return @worktime;
}

sub SubmitEntry($$$) {
    my ($self, $user, $task) = @_;

    my $result = undef;
    if ($self->Connect) {
        if ($task->GetID) {
            my $now = new Orangutan::Date;
            my $eventid = $task->GetID;
            my $projectid = $task->GetProject('id') || 0;
            my $activityid = $task->GetActivity('id') || 0;
            my $issueid = $task->GetIssue || 'NULL';
            my $hours = sprintf("%.2f", $task->GetDuration / 3600);
            my $spenton = $self->Escape($task->GetStart->WorkDayStart->DateTime('date'));
            my ($month, $year) = ($task->GetStart->GetDate)[1..2];
            my $week = $task->GetStart->GetWeek;
            my $reportdate = $self->Escape($now->DateTime);
            my $starttime = $self->Escape($task->GetStart->DateTime);
            my $endtime = $self->Escape($task->GetEnd->DateTime);
            if ($self->Query(qq{
                UPDATE time_entries
                SET
                    project_id  = $projectid,
                    issue_id    = $issueid,
                    hours       = $hours,
                    activity_id = $activityid,
                    spent_on    = $spenton,
                    tyear       = $year,
                    tmonth      = $month,
                    tweek       = $week,
                    updated_on  = $reportdate,
                    start_time  = $starttime,
                    end_time    = $endtime
                WHERE id = $eventid
                })) {
                $result = $eventid;
            }
        } else {
            my $now = new Orangutan::Date;
            my $projectid = $task->GetProject('id') || 0;
            my $activityid = $task->GetActivity('id') || 0;
            my $issueid = $task->GetIssue || 'NULL';
            my $employeeid = $user->GetUserID;
            my $starttime = $self->Escape($task->GetStart->DateTime);
            my $endtime = $self->Escape($task->GetEnd->DateTime);
            my $spenton = $self->Escape($task->GetStart->WorkDayStart->DateTime('date'));
            my ($month, $year) = ($task->GetStart->GetDate)[1..2];
            my $week = $task->GetStart->GetWeek;
            my $reportdate = $self->Escape($now->DateTime);
            my $hours = sprintf("%.2f", $task->GetDuration / 3600);
            my $description = $self->Escape($task->GetName);
            if ($self->Query(qq{
                INSERT INTO time_entries (
                    project_id,
                    user_id,
                    issue_id,
                    hours,
                    comments,
                    activity_id,
                    spent_on,
                    tyear,
                    tmonth,
                    tweek,
                    created_on,
                    updated_on,
                    start_time,
                    end_time
                )
                VALUES (
                    $projectid,
                    $employeeid,
                    $issueid,
                    $hours,
                    $description,
                    $activityid,
                    $spenton,
                    $year,
                    $month,
                    $week,
                    $reportdate,
                    $reportdate,
                    $starttime,
                    $endtime
                )})) {
                if ($self->Query(qq{
                    SELECT LAST_INSERT_ID() AS last_id
                    })) {
                    if ($self->Rows > 0) {
                        $result = $self->Next->{'last_id'};
                    }
                }
            }
        }
        if ($result) {
            my $chart = $self->GetChartEntries($result);
            if ($chart) {
                $self->UpdateChartEntries($chart);
            }
        }
        $self->Disconnect;
    }
    return $result;
}

sub RemoveEntry($$) {
    my ($self, $eventid) = @_;
    my $deleted = 0;
    if ($self->Connect) {
        my $chart = $self->GetChartEntries($eventid);
        if ($self->Query(qq{
            DELETE
            FROM time_entries
            WHERE id = $eventid
            })) {
            if ($chart) {
                $self->UpdateChartEntries($chart);
            }
            $deleted = 1;
        }
        $self->Disconnect;
    }
    return $deleted;
}

sub GetChartEntries($$) {
    my ($self, $eventid) = @_;

    my $result = undef;
    if ($self->Connect) {
        if ($self->Query(qq{
            SELECT chart.id, total.id AS total, spent_on,
            entries.user_id, entries.issue_id, entries.activity_id, entries.project_id
            FROM time_entries entries
            LEFT JOIN chart_time_entries chart
            ON chart.day = (YEAR(spent_on) * 1000 + DAYOFYEAR(spent_on))
            AND chart.user_id = entries.user_id
            AND chart.issue_id = entries.issue_id
            AND chart.activity_id = entries.activity_id
            AND chart.project_id = entries.project_id
            LEFT JOIN chart_time_entries total
            ON total.day = 0
            AND total.user_id = entries.user_id
            AND total.issue_id = entries.issue_id
            AND total.activity_id = entries.activity_id
            AND total.project_id = entries.project_id
            WHERE entries.id = $eventid
            })) {
            if ($self->Rows > 0) {
                $result = $self->Next;
            }
        }
        $self->Disconnect;
    }
    return $result;
}

sub UpdateChartEntries($$) {
    my ($self, $chart) = @_;

    my $updated = 0;
    if ($self->Connect) {

        if ($chart->{'id'}) {
            my $chartid = $chart->{'id'};
            my $spenton = $self->Escape($chart->{'spent_on'});
            my $employeeid = $chart->{'user_id'};
            my $issueid = $chart->{'issue_id'} || 'NULL';
            my $activityid = $chart->{'activity_id'} || 'NULL';
            my $projectid = $chart->{'project_id'};
            if ($self->Query(qq{
                SELECT
                    ROUND(SUM(hours), 2) AS logged_hours,
                    COUNT(id) AS entries
                FROM time_entries
                WHERE
                    spent_on = $spenton AND
                    user_id = $employeeid AND
                    issue_id = $issueid AND
                    activity_id = $activityid AND
                    project_id = $projectid
                GROUP BY spent_on, user_id, issue_id, activity_id, project_id
                })) {
                if ($self->Rows > 0) {
                    my $data = $self->Next;
                    my $loggedhours = $data->{'logged_hours'};
                    my $entries = $data->{'entries'};
                    if ($self->Query(qq{
                        UPDATE chart_time_entries
                        SET
                            logged_hours = $loggedhours,
                            entries = $entries
                        WHERE id = $chartid
                        })) {
                        $updated++;
                    }
                } else {
                    if ($self->Query(qq{
                        DELETE
                        FROM time_entries
                        WHERE id = $chartid
                        })) {
                        $updated++;
                    }
                }
            }
        } else {
            my $spenton = $self->Escape($chart->{'spent_on'});
            my $employeeid = $chart->{'user_id'};
            my $issueid = $chart->{'issue_id'} || 'NULL';
            my $activityid = $chart->{'activity_id'} || 'NULL';
            my $projectid = $chart->{'project_id'};
            if ($self->Query(qq{
                INSERT INTO chart_time_entries (
                    day,
                    week,
                    month,
                    logged_hours,
                    entries,
                    user_id,
                    issue_id,
                    activity_id,
                    project_id
                )
                SELECT
                    (YEAR(spent_on) * 1000 + DAYOFYEAR(spent_on)) AS day,
                    (YEAR(spent_on) * 1000 + WEEKOFYEAR(spent_on)) AS week,
                    (YEAR(spent_on) * 1000 + MONTH(spent_on)) AS month,
                    ROUND(SUM(hours), 2) AS logged_hours,
                    COUNT(id) AS entries,
                    user_id,
                    issue_id,
                    activity_id,
                    project_id
                FROM time_entries
                WHERE
                    spent_on = $spenton AND
                    user_id = $employeeid AND
                    issue_id = $issueid AND
                    activity_id = $activityid AND
                    project_id = $projectid
                GROUP BY spent_on, user_id, issue_id, activity_id, project_id
                })) {
                $updated++;
            }
        }

        if ($chart->{'total'}) {
            my $chartid = $chart->{'total'};
            my $employeeid = $chart->{'user_id'};
            my $issueid = $chart->{'issue_id'} || 'NULL';
            my $activityid = $chart->{'activity_id'} || 'NULL';
            my $projectid = $chart->{'project_id'};
            if ($self->Query(qq{
                SELECT
                    ROUND(SUM(hours), 2) AS logged_hours,
                    COUNT(id) AS entries
                FROM time_entries
                WHERE
                    user_id = $employeeid AND
                    issue_id = $issueid AND
                    activity_id = $activityid AND
                    project_id = $projectid
                GROUP BY user_id, issue_id, activity_id, project_id
                })) {
                if ($self->Rows > 0) {
                    my $data = $self->Next;
                    my $loggedhours = $data->{'logged_hours'};
                    my $entries = $data->{'entries'};
                    if ($self->Query(qq{
                        UPDATE chart_time_entries
                        SET
                            logged_hours = $loggedhours,
                            entries = $entries
                        WHERE id = $chartid
                        })) {
                        $updated++;
                    }
                } else {
                    if ($self->Query(qq{
                        DELETE
                        FROM time_entries
                        WHERE id = $chartid
                        })) {
                        $updated++;
                    }
                }
            }
        } else {
            my $employeeid = $chart->{'user_id'};
            my $issueid = $chart->{'issue_id'} || 'NULL';
            my $activityid = $chart->{'activity_id'} || 'NULL';
            my $projectid = $chart->{'project_id'};
            if ($self->Query(qq{
                INSERT INTO chart_time_entries (
                    day,
                    week,
                    month,
                    logged_hours,
                    entries,
                    user_id,
                    issue_id,
                    activity_id,
                    project_id
                )
                SELECT
                    0,
                    0,
                    0,
                    ROUND(SUM(hours), 2) AS logged_hours,
                    COUNT(id) AS entries,
                    user_id,
                    issue_id,
                    activity_id,
                    project_id
                FROM time_entries
                WHERE
                    user_id = $employeeid AND
                    issue_id = $issueid AND
                    activity_id = $activityid AND
                    project_id = $projectid
                GROUP BY user_id, issue_id, activity_id, project_id
                })) {
                $updated++;
            }
        }

        $self->Disconnect;
    }
    return $updated;
}

1;
